#! /usr/bin/env python
# encoding: utf-8

"""
waf-powered distributed network builds, with a network cache.

Caching files from a server has advantages over a NFS/Samba shared folder:

- builds are much faster because they use local files
- builds just continue to work in case of a network glitch
- permissions are much simpler to manage
"""

import os, urllib, tarfile, re, shutil, tempfile, sys
from collections import OrderedDict
from waflib import Context, Utils, Logs

try:
	from urllib.parse import urlencode
except ImportError:
	urlencode = urllib.urlencode

def safe_urlencode(data):
	x = urlencode(data)
	try:
		x = x.encode('utf-8')
	except Exception:
		pass
	return x

try:
	from urllib.error import URLError
except ImportError:
	from urllib2 import URLError

try:
	from urllib.request import Request, urlopen
except ImportError:
	from urllib2 import Request, urlopen

DISTNETCACHE = os.environ.get('DISTNETCACHE', '/tmp/distnetcache')
DISTNETSERVER = os.environ.get('DISTNETSERVER', 'http://localhost:8000/cgi-bin/')
TARFORMAT = 'w:bz2'
TIMEOUT = 60
REQUIRES = 'requires.txt'

re_com = re.compile(r'\s*#.*', re.M)

def total_version_order(num):
	lst = num.split('.')
	template = '%10s' * len(lst)
	ret = template % tuple(lst)
	return ret

def get_distnet_cache():
	return getattr(Context.g_module, 'DISTNETCACHE', DISTNETCACHE)

def get_server_url():
	return getattr(Context.g_module, 'DISTNETSERVER', DISTNETSERVER)

def get_download_url():
	return '%s/download.py' % get_server_url()

def get_upload_url():
	return '%s/upload.py' % get_server_url()

def get_resolve_url():
	return '%s/resolve.py' % get_server_url()

def send_package_name():
	out = getattr(Context.g_module, 'out', 'build')
	pkgfile = '%s/package_to_upload.tarfile' % out
	return pkgfile

class package(Context.Context):
	fun = 'package'
	cmd = 'package'

	def execute(self):
		try:
			files = self.files
		except AttributeError:
			files = self.files = []

		Context.Context.execute(self)
		pkgfile = send_package_name()
		if not pkgfile in files:
			if not REQUIRES in files:
				files.append(REQUIRES)
			self.make_tarfile(pkgfile, files, add_to_package=False)

	def make_tarfile(self, filename, files, **kw):
		if kw.get('add_to_package', True):
			self.files.append(filename)

		with tarfile.open(filename, TARFORMAT) as tar:
			endname = os.path.split(filename)[-1]
			endname = endname.split('.')[0] + '/'
			for x in files:
				tarinfo = tar.gettarinfo(x, x)
				tarinfo.uid   = tarinfo.gid   = 0
				tarinfo.uname = tarinfo.gname = 'root'
				tarinfo.size = os.stat(x).st_size
				if os.environ.get('SOURCE_DATE_EPOCH'):
					tarinfo.mtime = int(os.environ.get('SOURCE_DATE_EPOCH'))

				# TODO - more archive creation options?
				if kw.get('bare', True):
					tarinfo.name = os.path.split(x)[1]
				else:
					tarinfo.name = endname + x # todo, if tuple, then..
				Logs.debug('distnet: adding %r to %s', tarinfo.name, filename)
				with open(x, 'rb') as f:
					tar.addfile(tarinfo, f)
		Logs.info('Created %s', filename)

class publish(Context.Context):
	fun = 'publish'
	cmd = 'publish'
	def execute(self):
		if hasattr(Context.g_module, 'publish'):
			Context.Context.execute(self)
		mod = Context.g_module

		rfile = getattr(self, 'rfile', send_package_name())
		if not os.path.isfile(rfile):
			self.fatal('Create the release file with "waf release" first! %r' % rfile)

		fdata = Utils.readf(rfile, m='rb')
		data = safe_urlencode([('pkgdata', fdata), ('pkgname', mod.APPNAME), ('pkgver', mod.VERSION)])

		req = Request(get_upload_url(), data)
		response = urlopen(req, timeout=TIMEOUT)
		data = response.read().strip()

		if sys.hexversion>0x300000f:
			data = data.decode('utf-8')

		if data != 'ok':
			self.fatal('Could not publish the package %r' % data)

class constraint(object):
	def __init__(self, line=''):
		self.required_line = line
		self.info = []

		line = line.strip()
		if not line:
			return

		lst = line.split(',')
		if lst:
			self.pkgname = lst[0]
			self.required_version = lst[1]
			for k in lst:
				a, b, c = k.partition('=')
				if a and c:
					self.info.append((a, c))
	def __str__(self):
		buf = []
		buf.append(self.pkgname)
		buf.append(self.required_version)
		for k in self.info:
			buf.append('%s=%s' % k)
		return ','.join(buf)

	def __repr__(self):
		return "requires %s-%s" % (self.pkgname, self.required_version)

	def human_display(self, pkgname, pkgver):
		return '%s-%s requires %s-%s' % (pkgname, pkgver, self.pkgname, self.required_version)

	def why(self):
		ret = []
		for x in self.info:
			if x[0] == 'reason':
				ret.append(x[1])
		return ret

	def add_reason(self, reason):
		self.info.append(('reason', reason))

def parse_constraints(text):
	assert(text is not None)
	constraints = []
	text = re.sub(re_com, '', text)
	lines = text.splitlines()
	for line in lines:
		line = line.strip()
		if not line:
			continue
		constraints.append(constraint(line))
	return constraints

def list_package_versions(cachedir, pkgname):
	pkgdir = os.path.join(cachedir, pkgname)
	try:
		versions = os.listdir(pkgdir)
	except OSError:
		return []
	versions.sort(key=total_version_order)
	versions.reverse()
	return versions

class package_reader(Context.Context):
	cmd = 'solver'
	fun = 'solver'

	def __init__(self, **kw):
		Context.Context.__init__(self, **kw)

		self.myproject = getattr(Context.g_module, 'APPNAME', 'project')
		self.myversion = getattr(Context.g_module, 'VERSION', '1.0')
		self.cache_constraints = {}
		self.constraints = []

	def compute_dependencies(self, filename=REQUIRES):
		text = Utils.readf(filename)
		data = safe_urlencode([('text', text)])

		if '--offline' in sys.argv:
			self.constraints = self.local_resolve(text)
		else:
			req = Request(get_resolve_url(), data)
			try:
				response = urlopen(req, timeout=TIMEOUT)
			except URLError as e:
				Logs.warn('The package server is down! %r', e)
				self.constraints = self.local_resolve(text)
			else:
				ret = response.read()
				try:
					ret = ret.decode('utf-8')
				except Exception:
					pass
				self.trace(ret)
				self.constraints = parse_constraints(ret)
		self.check_errors()

	def check_errors(self):
		errors = False
		for c in self.constraints:
			if not c.required_version:
				errors = True

				reasons = c.why()
				if len(reasons) == 1:
					Logs.error('%s but no matching package could be found in this repository', reasons[0])
				else:
					Logs.error('Conflicts on package %r:', c.pkgname)
					for r in reasons:
						Logs.error('  %s', r)
		if errors:
			self.fatal('The package requirements cannot be satisfied!')

	def load_constraints(self, pkgname, pkgver, requires=REQUIRES):
		try:
			return self.cache_constraints[(pkgname, pkgver)]
		except KeyError:
			text = Utils.readf(os.path.join(get_distnet_cache(), pkgname, pkgver, requires))
			ret = parse_constraints(text)
			self.cache_constraints[(pkgname, pkgver)] = ret
			return ret

	def apply_constraint(self, domain, constraint):
		vname = constraint.required_version.replace('*', '.*')
		rev = re.compile(vname, re.M)
		ret = [x for x in domain if rev.match(x)]
		return ret

	def trace(self, *k):
		if getattr(self, 'debug', None):
			Logs.error(*k)

	def solve(self, packages_to_versions={}, packages_to_constraints={}, pkgname='', pkgver='', todo=[], done=[]):
		# breadth first search
		n_packages_to_versions = dict(packages_to_versions)
		n_packages_to_constraints = dict(packages_to_constraints)

		self.trace("calling solve with %r    %r %r" % (packages_to_versions, todo, done))
		done = done + [pkgname]

		constraints = self.load_constraints(pkgname, pkgver)
		self.trace("constraints %r" % constraints)

		for k in constraints:
			try:
				domain = n_packages_to_versions[k.pkgname]
			except KeyError:
				domain = list_package_versions(get_distnet_cache(), k.pkgname)


			self.trace("constraints?")
			if not k.pkgname in done:
				todo = todo + [k.pkgname]

			self.trace("domain before %s -> %s, %r" % (pkgname, k.pkgname, domain))

			# apply the constraint
			domain = self.apply_constraint(domain, k)

			self.trace("domain after %s -> %s, %r" % (pkgname, k.pkgname, domain))

			n_packages_to_versions[k.pkgname] = domain

			# then store the constraint applied
			constraints = list(packages_to_constraints.get(k.pkgname, []))
			constraints.append((pkgname, pkgver, k))
			n_packages_to_constraints[k.pkgname] = constraints

			if not domain:
				self.trace("no domain while processing constraint %r from %r %r" % (domain, pkgname, pkgver))
				return (n_packages_to_versions, n_packages_to_constraints)

		# next package on the todo list
		if not todo:
			return (n_packages_to_versions, n_packages_to_constraints)

		n_pkgname = todo[0]
		n_pkgver = n_packages_to_versions[n_pkgname][0]
		tmp = dict(n_packages_to_versions)
		tmp[n_pkgname] = [n_pkgver]

		self.trace("fixed point %s" % n_pkgname)

		return self.solve(tmp, n_packages_to_constraints, n_pkgname, n_pkgver, todo[1:], done)

	def get_results(self):
		return '\n'.join([str(c) for c in self.constraints])

	def solution_to_constraints(self, versions, constraints):
		solution = []
		for p in versions:
			c = constraint()
			solution.append(c)

			c.pkgname = p
			if versions[p]:
				c.required_version = versions[p][0]
			else:
				c.required_version = ''
			for (from_pkgname, from_pkgver, c2) in constraints.get(p, ''):
				c.add_reason(c2.human_display(from_pkgname, from_pkgver))
		return solution

	def local_resolve(self, text):
		self.cache_constraints[(self.myproject, self.myversion)] = parse_constraints(text)
		p2v = OrderedDict({self.myproject: [self.myversion]})
		(versions, constraints) = self.solve(p2v, {}, self.myproject, self.myversion, [])
		return self.solution_to_constraints(versions, constraints)

	def download_to_file(self, pkgname, pkgver, subdir, tmp):
		data = safe_urlencode([('pkgname', pkgname), ('pkgver', pkgver), ('pkgfile', subdir)])
		req = urlopen(get_download_url(), data, timeout=TIMEOUT)
		with open(tmp, 'wb') as f:
			while True:
				buf = req.read(8192)
				if not buf:
					break
				f.write(buf)

	def extract_tar(self, subdir, pkgdir, tmpfile):
		with tarfile.open(tmpfile) as f:
			temp = tempfile.mkdtemp(dir=pkgdir)
			try:
				if hasattr(tarfile, 'data_filter'):
					f.extractall(temp, filter='data')
				else:
					f.extractall(temp)
				os.rename(temp, os.path.join(pkgdir, subdir))
			finally:
				try:
					shutil.rmtree(temp)
				except Exception:
					pass

	def get_pkg_dir(self, pkgname, pkgver, subdir):
		pkgdir = os.path.join(get_distnet_cache(), pkgname, pkgver)
		if not os.path.isdir(pkgdir):
			os.makedirs(pkgdir)

		target = os.path.join(pkgdir, subdir)

		if os.path.exists(target):
			return target

		(fd, tmp) = tempfile.mkstemp(dir=pkgdir)
		try:
			os.close(fd)
			self.download_to_file(pkgname, pkgver, subdir, tmp)
			if subdir == REQUIRES:
				os.rename(tmp, target)
			else:
				self.extract_tar(subdir, pkgdir, tmp)
		finally:
			try:
				os.remove(tmp)
			except OSError:
				pass

		return target

	def __iter__(self):
		if not self.constraints:
			self.compute_dependencies()
		for x in self.constraints:
			if x.pkgname == self.myproject:
				continue
			yield x

	def execute(self):
		self.compute_dependencies()

packages = package_reader()

def load_tools(ctx, extra):
	global packages
	for c in packages:
		packages.get_pkg_dir(c.pkgname, c.required_version, extra)
		noarchdir = packages.get_pkg_dir(c.pkgname, c.required_version, 'noarch')
		for x in os.listdir(noarchdir):
			if x.startswith('waf_') and x.endswith('.py'):
				ctx.load([x.rstrip('.py')], tooldir=[noarchdir])

def options(opt):
	opt.add_option('--offline', action='store_true')
	packages.execute()
	load_tools(opt, REQUIRES)

def configure(conf):
	load_tools(conf, conf.variant)

def build(bld):
	load_tools(bld, bld.variant)




#! /usr/bin/env python
# per rosengren 2011

from os import sep, readlink
from waflib import Logs
from waflib.TaskGen import feature, after_method
from waflib.Task import Task, always_run

def options(opt):
	grp = opt.add_option_group('Bjam Options')
	grp.add_option('--bjam_src', default=None, help='You can find it in <boost root>/tools/jam/src')
	grp.add_option('--bjam_uname', default='linuxx86_64', help='bjam is built in <src>/bin.<uname>/bjam')
	grp.add_option('--bjam_config', default=None)
	grp.add_option('--bjam_toolset', default=None)

def configure(cnf):
	if not cnf.env.BJAM_SRC:
		cnf.env.BJAM_SRC = cnf.options.bjam_src
	if not cnf.env.BJAM_UNAME:
		cnf.env.BJAM_UNAME = cnf.options.bjam_uname
	try:
		cnf.find_program('bjam', path_list=[
			cnf.env.BJAM_SRC + sep + 'bin.' + cnf.env.BJAM_UNAME
		])
	except Exception:
		cnf.env.BJAM = None
	if not cnf.env.BJAM_CONFIG:
		cnf.env.BJAM_CONFIG = cnf.options.bjam_config
	if not cnf.env.BJAM_TOOLSET:
		cnf.env.BJAM_TOOLSET = cnf.options.bjam_toolset

@feature('bjam')
@after_method('process_rule')
def process_bjam(self):
	if not self.bld.env.BJAM:
		self.create_task('bjam_creator')
	self.create_task('bjam_build')
	self.create_task('bjam_installer')
	if getattr(self, 'always', False):
		always_run(bjam_creator)
		always_run(bjam_build)
	always_run(bjam_installer)

class bjam_creator(Task):
	ext_out = 'bjam_exe'
	vars=['BJAM_SRC', 'BJAM_UNAME']
	def run(self):
		env = self.env
		gen = self.generator
		bjam = gen.bld.root.find_dir(env.BJAM_SRC)
		if not bjam:
			Logs.error('Can not find bjam source')
			return -1
		bjam_exe_relpath = 'bin.' + env.BJAM_UNAME + '/bjam'
		bjam_exe = bjam.find_resource(bjam_exe_relpath)
		if bjam_exe:
			env.BJAM = bjam_exe.srcpath()
			return 0
		bjam_cmd = ['./build.sh']
		Logs.debug('runner: ' + bjam.srcpath() + '> ' + str(bjam_cmd))
		result = self.exec_command(bjam_cmd, cwd=bjam.srcpath())
		if not result == 0:
			Logs.error('bjam failed')
			return -1
		bjam_exe = bjam.find_resource(bjam_exe_relpath)
		if bjam_exe:
			env.BJAM = bjam_exe.srcpath()
			return 0
		Logs.error('bjam failed')
		return -1

class bjam_build(Task):
	ext_in = 'bjam_exe'
	ext_out = 'install'
	vars = ['BJAM_TOOLSET']
	def run(self):
		env = self.env
		gen = self.generator
		path = gen.path
		bld = gen.bld
		if hasattr(gen, 'root'):
			build_root = path.find_node(gen.root)
		else:
			build_root = path
		jam = bld.srcnode.find_resource(env.BJAM_CONFIG)
		if jam:
			Logs.debug('bjam: Using jam configuration from ' + jam.srcpath())
			jam_rel = jam.relpath_gen(build_root)
		else:
			Logs.warn('No build configuration in build_config/user-config.jam. Using default')
			jam_rel = None
		bjam_exe = bld.srcnode.find_node(env.BJAM)
		if not bjam_exe:
			Logs.error('env.BJAM is not set')
			return -1
		bjam_exe_rel = bjam_exe.relpath_gen(build_root)
		cmd = ([bjam_exe_rel] +
			(['--user-config=' + jam_rel] if jam_rel else []) +
			['--stagedir=' + path.get_bld().path_from(build_root)] +
			['--debug-configuration'] +
			['--with-' + lib for lib in self.generator.target] +
			(['toolset=' + env.BJAM_TOOLSET] if env.BJAM_TOOLSET else []) +
			['link=' + 'shared'] +
			['variant=' + 'release']
		)
		Logs.debug('runner: ' + build_root.srcpath() + '> ' + str(cmd))
		ret = self.exec_command(cmd, cwd=build_root.srcpath())
		if ret != 0:
			return ret
		self.set_outputs(path.get_bld().ant_glob('lib/*') + path.get_bld().ant_glob('bin/*'))
		return 0

class bjam_installer(Task):
	ext_in = 'install'
	def run(self):
		gen = self.generator
		path = gen.path
		for idir, pat in (('${LIBDIR}', 'lib/*'), ('${BINDIR}', 'bin/*')):
			files = []
			for n in path.get_bld().ant_glob(pat):
				try:
					t = readlink(n.srcpath())
					gen.bld.symlink_as(sep.join([idir, n.name]), t, postpone=False)
				except OSError:
					files.append(n)
			gen.bld.install_files(idir, files, postpone=False)
		return 0




"""Support for Sphinx documentation

This is a wrapper for sphinx-build program. Please note that sphinx-build supports only
one output format at a time, but the tool can create multiple tasks to handle more.
The output formats can be passed via the sphinx_output_format, which is an array of
strings. For backwards compatibility if only one output is needed, it can be passed
as a single string.
The default output format is html.

Specific formats can be installed in different directories by specifying the
install_path_<FORMAT> attribute. If not defined, the standard install_path
will be used instead.

Example wscript:

def configure(cnf):
    conf.load('sphinx')

def build(bld):
    bld(
        features='sphinx',
        sphinx_source='sources',  # path to source directory
        sphinx_options='-a -v',  # sphinx-build program additional options
        sphinx_output_format=['html', 'man'],  # output format of sphinx documentation
        install_path_man='${DOCDIR}/man'       # put man pages in a specific directory
        )

"""

from waflib.Node import Node
from waflib import Utils
from waflib import Task
from waflib.TaskGen import feature, after_method


def configure(cnf):
    """Check if sphinx-build program is available and loads gnu_dirs tool."""
    cnf.find_program('sphinx-build', var='SPHINX_BUILD', mandatory=False)
    cnf.load('gnu_dirs')


@feature('sphinx')
def build_sphinx(self):
    """Builds sphinx sources.
    """
    if not self.env.SPHINX_BUILD:
        self.bld.fatal('Program SPHINX_BUILD not defined.')
    if not getattr(self, 'sphinx_source', None):
        self.bld.fatal('Attribute sphinx_source not defined.')
    if not isinstance(self.sphinx_source, Node):
        self.sphinx_source = self.path.find_node(self.sphinx_source)
    if not self.sphinx_source:
        self.bld.fatal('Can\'t find sphinx_source: %r' % self.sphinx_source)

    # In the taskgen we have the complete list of formats
    Utils.def_attrs(self, sphinx_output_format='html')
    self.sphinx_output_format = Utils.to_list(self.sphinx_output_format)

    self.env.SPHINX_OPTIONS = getattr(self, 'sphinx_options', [])

    for source_file in self.sphinx_source.ant_glob('**/*'):
        self.bld.add_manual_dependency(self.sphinx_source, source_file)

    for cfmt in self.sphinx_output_format:
        sphinx_build_task = self.create_task('SphinxBuildingTask')
        sphinx_build_task.set_inputs(self.sphinx_source)
        # In task we keep the specific format this task is generating
        sphinx_build_task.env.SPHINX_OUTPUT_FORMAT = cfmt

        # the sphinx-build results are in <build + output_format> directory
        sphinx_build_task.sphinx_output_directory = self.path.get_bld().make_node(cfmt)
        sphinx_build_task.set_outputs(sphinx_build_task.sphinx_output_directory)
        sphinx_build_task.sphinx_output_directory.mkdir()

        Utils.def_attrs(sphinx_build_task, install_path=getattr(self, 'install_path_' + cfmt, getattr(self, 'install_path', get_install_path(sphinx_build_task))))


def get_install_path(object):
    if object.env.SPHINX_OUTPUT_FORMAT == 'man':
        return object.env.MANDIR
    elif object.env.SPHINX_OUTPUT_FORMAT == 'info':
        return object.env.INFODIR
    else:
        return object.env.DOCDIR


class SphinxBuildingTask(Task.Task):
    color = 'BOLD'
    run_str = '${SPHINX_BUILD} -M ${SPHINX_OUTPUT_FORMAT} ${SRC} ${TGT} -d ${TGT[0].bld_dir()}/doctrees-${SPHINX_OUTPUT_FORMAT} ${SPHINX_OPTIONS}'

    def keyword(self):
        return 'Compiling (%s)' % self.env.SPHINX_OUTPUT_FORMAT

    def runnable_status(self):

        for x in self.run_after:
            if not x.hasrun:
                return Task.ASK_LATER

        self.signature()
        ret = Task.Task.runnable_status(self)
        if ret == Task.SKIP_ME:
            # in case the files were removed
            self.add_install()
        return ret


    def post_run(self):
        self.add_install()
        return Task.Task.post_run(self)


    def add_install(self):
        nodes = self.sphinx_output_directory.ant_glob('**/*', quiet=True)
        self.outputs += nodes
        self.generator.add_install_files(install_to=self.install_path,
                                         install_from=nodes,
                                         postpone=False,
                                         cwd=self.sphinx_output_directory.make_node(self.env.SPHINX_OUTPUT_FORMAT),
                                         relative_trick=True)



#!/usr/bin/env python
# encoding: utf-8
# Thomas Nagy, 2010 (ita)

"""
Scala support

scalac outputs files a bit where it wants to
"""

import os
from waflib import Task, Utils, Node
from waflib.TaskGen import feature, before_method, after_method

from waflib.Tools import ccroot
ccroot.USELIB_VARS['scalac'] = set(['CLASSPATH', 'SCALACFLAGS'])

from waflib.Tools import javaw

@feature('scalac')
@before_method('process_source')
def apply_scalac(self):

	Utils.def_attrs(self, jarname='', classpath='',
		sourcepath='.', srcdir='.',
		jar_mf_attributes={}, jar_mf_classpath=[])

	outdir = getattr(self, 'outdir', None)
	if outdir:
		if not isinstance(outdir, Node.Node):
			outdir = self.path.get_bld().make_node(self.outdir)
	else:
		outdir = self.path.get_bld()
	outdir.mkdir()
	self.outdir = outdir
	self.env['OUTDIR'] = outdir.abspath()

	self.scalac_task = tsk = self.create_task('scalac')
	tmp = []

	srcdir = getattr(self, 'srcdir', '')
	if isinstance(srcdir, Node.Node):
		srcdir = [srcdir]
	for x in Utils.to_list(srcdir):
		if isinstance(x, Node.Node):
			y = x
		else:
			y = self.path.find_dir(x)
			if not y:
				self.bld.fatal('Could not find the folder %s from %s' % (x, self.path))
		tmp.append(y)
	tsk.srcdir = tmp

# reuse some code
feature('scalac')(javaw.use_javac_files)
after_method('apply_scalac')(javaw.use_javac_files)

feature('scalac')(javaw.set_classpath)
after_method('apply_scalac', 'use_scalac_files')(javaw.set_classpath)


SOURCE_RE = '**/*.scala'
class scalac(javaw.javac):
	color = 'GREEN'
	vars    = ['CLASSPATH', 'SCALACFLAGS', 'SCALAC', 'OUTDIR']

	def runnable_status(self):
		"""
		Wait for dependent tasks to be complete, then read the file system to find the input nodes.
		"""
		for t in self.run_after:
			if not t.hasrun:
				return Task.ASK_LATER

		if not self.inputs:
			global SOURCE_RE
			self.inputs  = []
			for x in self.srcdir:
				self.inputs.extend(x.ant_glob(SOURCE_RE, remove=False))
		return super(javaw.javac, self).runnable_status()

	def run(self):
		"""
		Execute the scalac compiler
		"""
		env = self.env
		gen = self.generator
		bld = gen.bld
		wd = bld.bldnode.abspath()
		def to_list(xx):
			if isinstance(xx, str):
				return [xx]
			return xx
		self.last_cmd = lst = []
		lst.extend(to_list(env['SCALAC']))
		lst.extend(['-classpath'])
		lst.extend(to_list(env['CLASSPATH']))
		lst.extend(['-d'])
		lst.extend(to_list(env['OUTDIR']))
		lst.extend(to_list(env['SCALACFLAGS']))
		lst.extend([a.abspath() for a in self.inputs])
		lst = [x for x in lst if x]
		try:
			self.out = self.generator.bld.cmd_and_log(lst, cwd=wd, env=env.env or None, output=0, quiet=0)[1]
		except:
			self.generator.bld.cmd_and_log(lst, cwd=wd, env=env.env or None)

def configure(self):
	"""
	Detect the scalac program
	"""
	# If SCALA_HOME is set, we prepend it to the path list
	java_path = self.environ['PATH'].split(os.pathsep)
	v = self.env

	if 'SCALA_HOME' in self.environ:
		java_path = [os.path.join(self.environ['SCALA_HOME'], 'bin')] + java_path
		self.env['SCALA_HOME'] = [self.environ['SCALA_HOME']]

	for x in 'scalac scala'.split():
		self.find_program(x, var=x.upper(), path_list=java_path)

	if 'CLASSPATH' in self.environ:
		v['CLASSPATH'] = self.environ['CLASSPATH']

	v.SCALACFLAGS = ['-verbose']
	if not v['SCALAC']:
		self.fatal('scalac is required for compiling scala classes')




import re

from waflib import Utils, Task, Errors, Logs
from waflib.Configure import conf
from waflib.TaskGen import extension, taskgen_method

HAXE_COMPILERS = {
    'JS': {'tgt': '--js', 'ext_out': ['.js']},
    'LUA': {'tgt': '--lua', 'ext_out': ['.lua']},
    'SWF': {'tgt': '--swf', 'ext_out': ['.swf']},
    'NEKO': {'tgt': '--neko', 'ext_out': ['.n']},
    'PHP': {'tgt': '--php', 'ext_out': ['.php']},
    'CPP': {'tgt': '--cpp', 'ext_out': ['.h', '.cpp']},
    'CPPIA': {'tgt': '--cppia', 'ext_out': ['.cppia']},
    'CS': {'tgt': '--cs', 'ext_out': ['.cs']},
    'JAVA': {'tgt': '--java', 'ext_out': ['.java']},
    'JVM': {'tgt': '--jvm', 'ext_out': ['.jar']},
    'PYTHON': {'tgt': '--python', 'ext_out': ['.py']},
    'HL': {'tgt': '--hl', 'ext_out': ['.hl']},
    'HLC': {'tgt': '--hl', 'ext_out': ['.h', '.c']},
}

@conf
def check_haxe_pkg(self, **kw):
    self.find_program('haxelib')
    libs = kw.get('libs')
    if not libs or not (type(libs) == str or (type(libs) == list and all(isinstance(s, str) for s in libs))):
        self.fatal('Specify correct libs value in ensure call')
        return
    fetch = kw.get('fetch')
    if not fetch is None and not type(fetch) == bool:
        self.fatal('Specify correct fetch value in ensure call')

    libs = [libs] if type(libs) == str else libs
    halt = False
    for lib in libs:
        try:
            self.start_msg('Checking for library %s' % lib)
            output = self.cmd_and_log(self.env.HAXELIB + ['list', lib])
        except Errors.WafError:
            self.end_msg(False)
            self.fatal('Can\'t run haxelib list, ensuring halted')
            return

        if lib in output:
            self.end_msg(lib in output)
        else:
            if not fetch:
                self.end_msg(False)
                halt = True
                continue
            try:
                status = self.exec_command(self.env.HAXELIB + ['install', lib])
                if status:
                    self.end_msg(False)
                    self.fatal('Can\'t get %s with haxelib, ensuring halted' % lib)
                    return
                else:
                    self.end_msg('downloaded', color='YELLOW')
            except Errors.WafError:
                self.end_msg(False)
                self.fatal('Can\'t run haxelib install, ensuring halted')
                return
        postfix = kw.get('uselib_store') or lib.upper()
        self.env.append_unique('LIB_' + postfix, lib)

    if halt:
        self.fatal('Can\'t find libraries in haxelib list, ensuring halted')
        return

class haxe(Task.Task):
    vars = ['HAXE_VERSION', 'HAXE_FLAGS']
    ext_in = ['.hx']

    def run(self):
        cmd = self.env.HAXE + self.env.HAXE_FLAGS_DEFAULT + self.env.HAXE_FLAGS
        return self.exec_command(cmd)

for COMP in HAXE_COMPILERS:
    # create runners for each compile target
    type("haxe_" + COMP, (haxe,), {'ext_out': HAXE_COMPILERS[COMP]['ext_out']})

@taskgen_method
def init_haxe(self):
    errmsg = '%s not found, specify correct value'
    try:
        compiler = HAXE_COMPILERS[self.compiler]
        comp_tgt = compiler['tgt']
        comp_mod = '/main.c' if self.compiler == 'HLC' else ''
    except (AttributeError, KeyError):
        self.bld.fatal(errmsg % 'COMPILER' + ': ' + ', '.join(HAXE_COMPILERS.keys()))
        return

    self.env.append_value(
        'HAXE_FLAGS',
        [comp_tgt, self.path.get_bld().make_node(self.target + comp_mod).abspath()])
    if hasattr(self, 'use'):
        if not (type(self.use) == str or type(self.use) == list):
            self.bld.fatal(errmsg % 'USE')
            return
        self.use = [self.use] if type(self.use) == str else self.use

        for dep in self.use:
            if self.env['LIB_' + dep]:
                for lib in self.env['LIB_' + dep]:
                    self.env.append_value('HAXE_FLAGS', ['-lib', lib])

    if hasattr(self, 'res'):
        if not type(self.res) == str:
            self.bld.fatal(errmsg % 'RES')
            return
        self.env.append_value('HAXE_FLAGS', ['-D', 'resourcesPath=%s' % self.res])

@extension('.hx')
def haxe_hook(self, node):
    if len(self.source) > 1:
        self.bld.fatal('Use separate task generators for multiple files')
        return

    src = node
    tgt = self.path.get_bld().find_or_declare(self.target)

    self.init_haxe()
    self.create_task('haxe_' + self.compiler, src, tgt)

@conf
def check_haxe(self, mini=None, maxi=None):
    self.start_msg('Checking for haxe version')
    try:
        curr = re.search(
            r'(\d+.?)+',
            self.cmd_and_log(self.env.HAXE + ['-version'])).group()
    except Errors.WafError:
        self.end_msg(False)
        self.fatal('Can\'t get haxe version')
        return

    if mini and Utils.num2ver(curr) < Utils.num2ver(mini):
        self.end_msg('wrong', color='RED')
        self.fatal('%s is too old, need >= %s' % (curr, mini))
        return
    if maxi and Utils.num2ver(curr) > Utils.num2ver(maxi):
        self.end_msg('wrong', color='RED')
        self.fatal('%s is too new, need <= %s' % (curr, maxi))
        return
    self.end_msg(curr, color='GREEN')
    self.env.HAXE_VERSION = curr

def configure(self):
    self.env.append_value(
        'HAXE_FLAGS_DEFAULT',
        ['-D', 'no-compilation', '-cp', self.path.abspath()])
    Logs.warn('Default flags: %s' % ' '.join(self.env.HAXE_FLAGS_DEFAULT))
    self.find_program('haxe')



#! /usr/bin/env python
# encoding: utf-8
# KWS 2010
# Thomas Nagy 2016-2018 (ita)

import re
from waflib import Utils
from waflib.Tools import fc, fc_config, fc_scan, ar
from waflib.Configure import conf

@conf
def find_g95(conf):
	fc = conf.find_program('g95', var='FC')
	conf.get_g95_version(fc)
	conf.env.FC_NAME = 'G95'

@conf
def g95_flags(conf):
	v = conf.env
	v.FCFLAGS_fcshlib   = ['-fPIC']
	v.FORTRANMODFLAG  = ['-fmod=', ''] # template for module path
	v.FCFLAGS_DEBUG = ['-Werror'] # why not

@conf
def g95_modifier_win32(conf):
	fc_config.fortran_modifier_win32(conf)

@conf
def g95_modifier_cygwin(conf):
	fc_config.fortran_modifier_cygwin(conf)

@conf
def g95_modifier_darwin(conf):
	fc_config.fortran_modifier_darwin(conf)

@conf
def g95_modifier_platform(conf):
	dest_os = conf.env.DEST_OS or Utils.unversioned_sys_platform()
	g95_modifier_func = getattr(conf, 'g95_modifier_' + dest_os, None)
	if g95_modifier_func:
		g95_modifier_func()

@conf
def get_g95_version(conf, fc):
	"""get the compiler version"""

	version_re = re.compile(r"g95\s*(?P<major>\d*)\.(?P<minor>\d*)").search
	cmd = fc + ['--version']
	out, err = fc_config.getoutput(conf, cmd, stdin=False)
	if out:
		match = version_re(out)
	else:
		match = version_re(err)
	if not match:
		conf.fatal('cannot determine g95 version')
	k = match.groupdict()
	conf.env.FC_VERSION = (k['major'], k['minor'])

def configure(conf):
	conf.find_g95()
	conf.find_ar()
	conf.fc_flags()
	conf.fc_add_flags()
	conf.g95_flags()
	conf.g95_modifier_platform()




#! /usr/bin/env python
# encoding: utf-8

"""
Waf 1.6

Try to detect if the project directory was relocated, and if it was,
change the node representing the project directory. Just call:

 waf configure build

Note that if the project directory name changes, the signatures for the tasks using
files in that directory will change, causing a partial build.
"""

import os
from waflib import Build, ConfigSet, Task, Utils, Errors
from waflib.TaskGen import feature, after_method

EXTRA_LOCK = '.old_srcdir'

old1 = Build.BuildContext.store
def store(self):
	old1(self)
	db = os.path.join(self.variant_dir, EXTRA_LOCK)
	env = ConfigSet.ConfigSet()
	env.SRCDIR = self.srcnode.abspath()
	env.store(db)
Build.BuildContext.store = store

old2 = Build.BuildContext.init_dirs
def init_dirs(self):

	if not (os.path.isabs(self.top_dir) and os.path.isabs(self.out_dir)):
		raise Errors.WafError('The project was not configured: run "waf configure" first!')

	srcdir = None
	db = os.path.join(self.variant_dir, EXTRA_LOCK)
	env = ConfigSet.ConfigSet()
	try:
		env.load(db)
		srcdir = env.SRCDIR
	except:
		pass

	if srcdir:
		d = self.root.find_node(srcdir)
		if d and srcdir != self.top_dir and getattr(d, 'children', ''):
			srcnode = self.root.make_node(self.top_dir)
			print("relocating the source directory %r -> %r" % (srcdir, self.top_dir))
			srcnode.children = {}

			for (k, v) in d.children.items():
				srcnode.children[k] = v
				v.parent = srcnode
			d.children = {}

	old2(self)

Build.BuildContext.init_dirs = init_dirs


def uid(self):
	try:
		return self.uid_
	except AttributeError:
		# this is not a real hot zone, but we want to avoid surprises here
		m = Utils.md5(usedforsecurity=False)
		up = m.update
		up(self.__class__.__name__.encode())
		for x in self.inputs + self.outputs:
			up(x.path_from(x.ctx.srcnode).encode())
		self.uid_ = m.digest()
		return self.uid_
Task.Task.uid = uid

@feature('c', 'cxx', 'd', 'go', 'asm', 'fc', 'includes')
@after_method('propagate_uselib_vars', 'process_source')
def apply_incpaths(self):
	lst = self.to_incnodes(self.to_list(getattr(self, 'includes', [])) + self.env['INCLUDES'])
	self.includes_nodes = lst
	bld = self.bld
	self.env['INCPATHS'] = [x.is_child_of(bld.srcnode) and x.path_from(bld.bldnode) or x.abspath() for x in lst]


